# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import numpy as np
from numpy.testing import assert_allclose

from MDAnalysisTests.datafiles import PDBX
from MDAnalysisTests.coordinates.reference import RefAdKSmall

from MDAnalysisTests.coordinates.base import _SingleFrameReader

import MDAnalysis as mda

try:
    import openmm as mm
    from openmm import unit, app
except ImportError:
    try:
        from simtk import openmm as mm
        from simtk import unit
        from simtk.openmm import app
    except ImportError:
        pytest.skip(allow_module_level=True)


class TestOpenMMBasicSimulationReader:
    @pytest.fixture
    def omm_sim_uni(self):
        system = mm.System()
        topology = app.Topology()
        chain = topology.addChain("CHAIN")
        hydrogen = app.element.Element.getByAtomicNumber(1)
        residue = topology.addResidue("RES", chain)
        for i in range(5):
            system.addParticle(1.0)
            topology.addAtom(hydrogen.symbol, hydrogen, residue)
        positions = np.ones((5, 3)) * unit.angstrom
        integrator = mm.LangevinIntegrator(
            273 * unit.kelvin,
            1.0 / unit.picoseconds,
            2.0 * unit.femtoseconds,
        )
        simulation = app.Simulation(topology, system, integrator)
        simulation.context.setPositions(positions)

        return mda.Universe(simulation)

    def test_dimensions(self, omm_sim_uni):
        assert_allclose(
            omm_sim_uni.trajectory.ts.dimensions,
            np.array([20.0, 20.0, 20.0, 90.0, 90.0, 90.0]),
            rtol=0,
            atol=1e-3,
            err_msg=(
                "OpenMMBasicSimulationReader failed to get unitcell "
                "dimensions from OpenMM Simulation Object"
            ),
        )

    def test_coordinates(self, omm_sim_uni):
        up = omm_sim_uni.atoms.positions
        reference = np.ones((5, 3))
        assert_allclose(up, reference, rtol=0, atol=1e-3)

    def test_basic_topology(self, omm_sim_uni):
        assert omm_sim_uni.atoms.n_atoms == 5
        assert omm_sim_uni.residues.n_residues == 1
        assert omm_sim_uni.residues.resnames[0] == "RES"
        assert omm_sim_uni.segments.n_segments == 1
        assert omm_sim_uni.segments.segids[0] == "0"
        assert len(omm_sim_uni.bonds.indices) == 0

    def test_data(self, omm_sim_uni):
        data = omm_sim_uni.trajectory.ts.data
        assert isinstance(data["kinetic_energy"], float)
        assert isinstance(data["potential_energy"], float)


class TestOpenMMPDBFileReader(_SingleFrameReader):
    __test__ = True

    def setUp(self):
        self.universe = mda.Universe(app.PDBFile(RefAdKSmall.filename))
        self.ref = mda.Universe(RefAdKSmall.filename)
        self.prec = 3

    def test_dimensions(self):
        assert_allclose(
            self.universe.trajectory.ts.dimensions,
            self.ref.trajectory.ts.dimensions,
            rtol=0,
            atol=1e-3,
            err_msg=(
                "OpenMMPDBFileReader failed to get unitcell dimensions "
                "from OpenMMPDBFile"
            ),
        )

    def test_coordinates(self):
        up = self.universe.atoms.positions
        rp = self.ref.atoms.positions
        assert_allclose(up, rp, rtol=0, atol=1e-3)


class TestOpenMMModellerReader(_SingleFrameReader):
    __test__ = True

    def setUp(self):
        pdb_obj = app.PDBFile(RefAdKSmall.filename)
        modeller = app.Modeller(pdb_obj.topology, pdb_obj.positions)
        self.universe = mda.Universe(modeller)
        self.ref = mda.Universe(RefAdKSmall.filename)
        self.prec = 3

    def test_dimensions(self):
        assert_allclose(
            self.universe.trajectory.ts.dimensions,
            self.ref.trajectory.ts.dimensions,
            rtol=0,
            atol=1e-3,
            err_msg=(
                "OpenMMModellerReader failed to get unitcell dimensions "
                "from OpenMMModeller"
            ),
        )

    def test_coordinates(self):
        up = self.universe.atoms.positions
        rp = self.ref.atoms.positions
        assert_allclose(up, rp, rtol=0, atol=1e-3)


class TestOpenMMSimulationReader(_SingleFrameReader):
    __test__ = True

    def setUp(self):
        pdb = app.PDBFile(RefAdKSmall.filename)
        forcefield = app.ForceField("amber99sbildn.xml")
        system = forcefield.createSystem(
            pdb.topology, nonbondedMethod=app.NoCutoff, constraints=app.HBonds
        )
        integrator = mm.LangevinIntegrator(
            300 * unit.kelvin, 1 / unit.picoseconds, 2.0 * unit.femtoseconds
        )
        sim = app.Simulation(pdb.topology, system, integrator)
        sim.context.setPositions(pdb.positions)
        self.universe = mda.Universe(sim)
        self.ref = mda.Universe(RefAdKSmall.filename)
        self.prec = 3

    def test_dimensions(self):
        assert_allclose(
            self.universe.trajectory.ts.dimensions,
            self.ref.trajectory.ts.dimensions,
            rtol=0,
            atol=1e-3,
            err_msg=(
                "OpenMMSimulationReader failed to get unitcell "
                "dimensions from OpenMMSimulation"
            ),
        )

    def test_coordinates(self):
        up = self.universe.atoms.positions
        rp = self.ref.atoms.positions
        assert_allclose(up, rp, rtol=0, atol=1e-3)

    @pytest.mark.xfail(reason="OpenMM pickling not supported yet")
    def test_pickle_singleframe_reader(self):
        """
        See `OpenMM SwigPyObject serialisation discussion <https://github.com/MDAnalysis/mdanalysis/issues/2887>`_
        """
        super().test_pickle_singleframe_reader()


@pytest.fixture
def PDBX_U():
    return mda.Universe(app.PDBxFile(PDBX))


def test_pdbx_coordinates(PDBX_U):
    ref_pos = 10 * np.array(
        [
            [0.6537, 3.7168, 1.4751],
            [0.6141, 3.7813, 1.3496],
            [0.4851, 3.8633, 1.3566],
            [0.4428, 3.914, 1.2532],
            [0.6074, 3.6749, 1.2371],
            [0.7366, 3.5948, 1.2334],
            [0.8485, 3.6434, 1.1667],
            [0.7494, 3.477, 1.3064],
            [0.9699, 3.5741, 1.1709],
            [0.8706, 3.4079, 1.3107],
            [0.9803, 3.4575, 1.2437],
            [0.4265, 3.8799, 1.476],
            [0.3019, 3.9561, 1.4964],
            [0.3076, 4.1011, 1.4488],
            [0.2061, 4.1541, 1.4068],
            [0.2633, 3.9555, 1.6453],
            [0.1999, 3.8275, 1.6944],
            [0.1911, 3.7306, 1.6147],
            [0.1619, 3.8224, 1.8133],
            [0.424, 4.1656, 1.4575],
            [0.4388, 4.3071, 1.4222],
            [0.5294, 4.3324, 1.3023],
            [0.5726, 4.4457, 1.2828],
            [0.4853, 4.387, 1.5449],
            [0.3731, 4.4096, 1.6449],
            [0.4257, 4.4404, 1.7833],
            [0.3124, 4.4673, 1.8792],
            [0.5552, 4.2303, 1.2196],
            [0.6382, 4.2504, 1.1005],
            [0.555, 4.3159, 0.9885],
            [0.4442, 4.2696, 0.9581],
            [0.7097, 4.119, 1.0495],
            [0.7974, 4.0487, 1.1586],
            [0.7916, 4.1436, 0.9212],
            [0.9078, 4.133, 1.2278],
            [0.6112, 4.4208, 0.9256],
            [0.5533, 4.4859, 0.8069],
            [0.6548, 4.4763, 0.6937],
            [0.6177, 4.4543, 0.5787],
            [0.5139, 4.6313, 0.8343],
            [0.3789, 4.6433, 0.904],
            [0.7844, 4.4865, 0.7289],
            [0.8992, 4.4777, 0.6384],
            [0.9414, 4.3301, 0.6208],
            [1.0515, 4.2927, 0.6608],
            [1.0151, 4.5643, 0.6923],
            [0.8528, 4.2473, 0.5599],
            [0.8763, 4.1044, 0.5361],
            [1.0028, 4.0746, 0.4537],
            [1.0652, 3.9719, 0.478],
            [0.7534, 4.0355, 0.4762],
            [0.6391, 4.0245, 0.5727],
            [0.5337, 4.1106, 0.5851],
            [0.6204, 3.9232, 0.6731],
            [0.4486, 4.0675, 0.6847],
            [0.5002, 3.9534, 0.7413],
            [0.695, 3.8109, 0.714],
            [0.4534, 3.8758, 0.8486],
            [0.6468, 3.7325, 0.8175],
            [0.5277, 3.7648, 0.8837],
        ]
    )
    rp = PDBX_U.atoms.positions
    assert_allclose(ref_pos, rp, rtol=0, atol=1e-3)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
ParmEd topology parser --- :mod:`MDAnalysis.converters.ParmEdParser`
====================================================================

Converts a `ParmEd <https://parmed.github.io/ParmEd/html>`_
:class:`parmed.structure.Structure` into a :class:`MDAnalysis.core.Topology`.


Example
-------

If you want to use an MDAnalysis-written ParmEd structure for simulation
in ParmEd, you need to first read your files with ParmEd to include the
necessary topology parameters. ::

    >>> import parmed as pmd
    >>> import MDAnalysis as mda
    >>> from MDAnalysis.tests.datafiles import PRM7_ala2, RST7_ala2
    >>> prm = pmd.load_file(PRM7_ala2, RST7_ala2)
    >>> prm
    <AmberParm 3026 atoms; 1003 residues; 3025 bonds; PBC (orthogonal); parametrized>

We can then convert this to an MDAnalysis structure, select only the
protein atoms, and then convert it back to ParmEd. ::

    >>> u = mda.Universe(prm)
    >>> u
    <Universe with 3026 atoms>
    >>> prot = u.select_atoms('protein')
    >>> prm_prot = prot.convert_to('PARMED')
    >>> prm_prot
    <Structure 23 atoms; 2 residues; 22 bonds; PBC (orthogonal); parametrized>

From here you can create an OpenMM simulation system and minimize the
energy. ::

    >>> import openmm as mm
    >>> import openmm.app as app
    >>> from parmed import unit as u
    >>> system = prm_prot.createSystem(nonbondedMethod=app.NoCutoff,
    ...                                constraints=app.HBonds,
    ...                                implicitSolvent=app.GBn2)
    >>> integrator = mm.LangevinIntegrator(
    ...                         300*u.kelvin,       # Temperature of heat bath
    ...                         1.0/u.picoseconds,  # Friction coefficient
    ...                         2.0*u.femtoseconds, # Time step
    ... )
    >>> sim = app.Simulation(prm_prot.topology, system, integrator)
    >>> sim.context.setPositions(prm_prot.positions)
    >>> sim.minimizeEnergy(maxIterations=500)

Now you can continue on and run a simulation, if you wish.

Classes
-------

.. autoclass:: ParmEdParser
   :members:
   :inherited-members:

.. versionchanged:: 2.0.0
   The ParmEdParser class was moved from :mod:`~MDAnalysis.topology` to
   :mod:`~MDAnalysis.converters`

"""
import logging
import numpy as np

from ..topology.base import TopologyReaderBase, change_squash
from ..guesser.tables import Z2SYMB
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    AltLocs,
    ChainIDs,
    Atomtypes,
    Occupancies,
    Tempfactors,
    Elements,
    Masses,
    Charges,
    Resids,
    Resnums,
    Resnames,
    Segids,
    GBScreens,
    SolventRadii,
    NonbondedIndices,
    RMins,
    Epsilons,
    RMin14s,
    Epsilon14s,
    Bonds,
    UreyBradleys,
    Angles,
    Dihedrals,
    Impropers,
    CMaps,
)
from ..core.topology import Topology

logger = logging.getLogger("MDAnalysis.converters.ParmEdParser")


def squash_identical(values):
    if len(values) == 1:
        return values[0]
    else:
        return tuple(values)


class ParmEdParser(TopologyReaderBase):
    """
    For ParmEd structures
    """

    format = "PARMED"

    @staticmethod
    def _format_hint(thing):
        """Can this Parser read object *thing*?

        .. versionadded:: 1.0.0
        """
        try:
            import parmed as pmd
        except ImportError:  # if no parmed, probably not parmed
            return False
        else:
            return isinstance(thing, pmd.Structure)

    def parse(self, **kwargs):
        """Parse PARMED into Topology

        Returns
        -------
        MDAnalysis *Topology* object


        .. versionchanged:: 2.0.0
           Elements are no longer guessed, if the elements present in the
           parmed object are not recoginsed (usually given an atomic mass of 0)
           then they will be assigned an empty string.
        """
        structure = self.filename

        #### === ATOMS === ####
        names = []
        masses = []
        charges = []
        types = []
        atomic_numbers = []
        serials = []

        resnames = []
        resids = []
        chainids = []
        segids = []

        altLocs = []
        bfactors = []
        occupancies = []

        screens = []
        solvent_radii = []
        nonbonded_indices = []

        rmins = []
        epsilons = []
        rmin14s = []
        epsilon14s = []

        for atom in structure.atoms:
            names.append(atom.name)
            masses.append(atom.mass)
            charges.append(atom.charge)
            types.append(atom.type)
            atomic_numbers.append(atom.atomic_number)
            serials.append(atom.number)

            resnames.append(atom.residue.name)
            resids.append(atom.residue.number)
            chainids.append(atom.residue.chain)
            segids.append(atom.residue.segid)

            altLocs.append(atom.altloc)
            bfactors.append(atom.bfactor)
            occupancies.append(atom.occupancy)

            screens.append(atom.screen)
            solvent_radii.append(atom.solvent_radius)
            nonbonded_indices.append(atom.nb_idx)

            rmins.append(atom.rmin)
            epsilons.append(atom.epsilon)
            rmin14s.append(atom.rmin_14)
            epsilon14s.append(atom.epsilon_14)

        attrs = []

        n_atoms = len(names)

        elements = []

        for z, name in zip(atomic_numbers, names):
            try:
                elements.append(Z2SYMB[z])
            except KeyError:
                elements.append("")

        # Make Atom TopologyAttrs
        for vals, Attr, dtype in (
            (names, Atomnames, object),
            (masses, Masses, np.float32),
            (charges, Charges, np.float32),
            (types, Atomtypes, object),
            (elements, Elements, object),
            (serials, Atomids, np.int32),
            (chainids, ChainIDs, object),
            (altLocs, AltLocs, object),
            (bfactors, Tempfactors, np.float32),
            (occupancies, Occupancies, np.float32),
            (screens, GBScreens, np.float32),
            (solvent_radii, SolventRadii, np.float32),
            (nonbonded_indices, NonbondedIndices, np.int32),
            (rmins, RMins, np.float32),
            (epsilons, Epsilons, np.float32),
            (rmin14s, RMin14s, np.float32),
            (epsilon14s, Epsilon14s, np.float32),
        ):
            attrs.append(Attr(np.array(vals, dtype=dtype)))

        resids = np.array(resids, dtype=np.int32)
        resnames = np.array(resnames, dtype=object)
        chainids = np.array(chainids, dtype=object)
        segids = np.array(segids, dtype=object)

        residx, (resids, resnames, chainids, segids) = change_squash(
            (resids, resnames, chainids, segids),
            (resids, resnames, chainids, segids),
        )

        n_residues = len(resids)
        attrs.append(Resids(resids))
        attrs.append(Resnums(resids.copy()))
        attrs.append(Resnames(resnames))

        segidx, (segids,) = change_squash((segids,), (segids,))
        n_segments = len(segids)
        attrs.append(Segids(segids))

        #### === OTHERS === ####
        bond_values = {}
        bond_types = []
        bond_orders = []

        ub_values = {}
        ub_types = []

        angle_values = {}
        angle_types = []

        dihedral_values = {}
        dihedral_types = []

        improper_values = {}
        improper_types = []

        cmap_values = {}
        cmap_types = []

        for bond in structure.bonds:
            idx = (bond.atom1.idx, bond.atom2.idx)
            if idx not in bond_values:
                bond_values[idx] = ([bond], [bond.order])
            else:
                bond_values[idx][0].append(bond)
                bond_values[idx][1].append(bond.order)

        try:
            bond_values, values = zip(*list(bond_values.items()))
        except ValueError:
            bond_values, bond_types, bond_orders = [], [], []
        else:
            bond_types, bond_orders = zip(*values)

            bond_types = list(map(squash_identical, bond_types))
            bond_orders = list(map(squash_identical, bond_orders))

        attrs.append(
            Bonds(
                bond_values, types=bond_types, guessed=False, order=bond_orders
            )
        )

        for pmdlist, na, values, types in (
            (structure.urey_bradleys, 2, ub_values, ub_types),
            (structure.angles, 3, angle_values, angle_types),
            (structure.dihedrals, 4, dihedral_values, dihedral_types),
            (structure.impropers, 4, improper_values, improper_types),
            (structure.cmaps, 5, cmap_values, cmap_types),
        ):

            for p in pmdlist:
                atoms = ["atom{}".format(i) for i in range(1, na + 1)]
                idx = tuple(getattr(p, a).idx for a in atoms)
                if idx not in values:
                    values[idx] = [p]
                else:
                    values[idx].append(p)

        for dct, Attr in (
            (ub_values, UreyBradleys),
            (angle_values, Angles),
            (dihedral_values, Dihedrals),
            (improper_values, Impropers),
            (cmap_values, CMaps),
        ):
            try:
                vals, types = zip(*list(dct.items()))
            except ValueError:
                vals, types = [], []

            types = list(map(squash_identical, types))
            attrs.append(Attr(vals, types=types, guessed=False, order=None))

        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            attrs=attrs,
            atom_resindex=residx,
            residue_segindex=segidx,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

# Test the selection exporters in MDAnalysis.selections
# use StringIO and NamedStream to write to memory instead to temp files
import pytest
from io import StringIO
from unittest.mock import patch

import re

import numpy as np
from numpy.testing import assert_equal, assert_array_equal

from MDAnalysis.tests.datafiles import PSF, DCD

import MDAnalysis
import MDAnalysis.selections as selections
from MDAnalysis.lib.util import NamedStream


class _SelectionWriter(object):

    filename = None
    max_number = (
        357  # to keep fixtures smallish, only select CAs up to number 357
    )

    @staticmethod
    @pytest.fixture()
    def universe():
        return MDAnalysis.Universe(PSF, DCD)

    @pytest.fixture()
    def namedfile(self):
        return NamedStream(StringIO(), self.filename)

    def _selection(self, universe):
        return universe.select_atoms(
            "protein and name CA and bynum 1-{0}".format(self.max_number)
        )

    def _write(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        g.write(namedfile, **kwargs)
        return g

    def _write_selection(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        g.write(namedfile, **kwargs)
        return g

    def _write_with(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        with self.writer(namedfile, **kwargs) as outfile:
            outfile.write(g)
        return g

    def test_write_bad_mode(
        self,
        universe,
        namedfile,
    ):
        with pytest.raises(ValueError):
            self._write(universe, namedfile, name=self.ref_name, mode="a+")

    def test_write(self, universe, namedfile):
        self._write(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)

    def test_writeselection(self, universe, namedfile):
        self._write_selection(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)

    def test_write_with(self, universe, namedfile):
        self._write_with(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)


def ndx2array(lines):
    """Convert Gromacs NDX text file lines to integer array"""
    return np.array(" ".join(lines).replace("\n", "").split(), dtype=int)


def lines2one(lines):
    """Join lines and squash all whitespace"""
    return " ".join(" ".join(lines).split())


class TestSelectionWriter_Gromacs(_SelectionWriter):

    writer = MDAnalysis.selections.gromacs.SelectionWriter
    filename = "CA.ndx"
    ref_name = "CA_selection"
    ref_indices = ndx2array(
        [
            "5 22 46 65 84 103 122 129 141 153 160 170 \n",
            "177 199 206 220 237 247 264 284 303 320 335 357 \n",
        ]
    )

    def _assert_selectionstring(self, namedfile):
        header = namedfile.readline().strip()
        assert_equal(
            header,
            "[ {0} ]".format(self.ref_name),
            err_msg="NDX file has wrong selection name",
        )
        indices = ndx2array(namedfile.readlines())
        assert_array_equal(
            indices,
            self.ref_indices,
            err_msg="indices were not written correctly",
        )


class TestSelectionWriter_Charmm(_SelectionWriter):

    writer = MDAnalysis.selections.charmm.SelectionWriter
    filename = "CA.str"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one(
        [
            """! MDAnalysis CHARMM selection
           DEFINE CA_selection SELECT -
           BYNUM 5 .or. BYNUM 22 .or. BYNUM 46 .or. BYNUM 65 .or. -
           BYNUM 84 .or. BYNUM 103 .or. BYNUM 122 .or. BYNUM 129 .or. -
           BYNUM 141 .or. BYNUM 153 .or. BYNUM 160 .or. BYNUM 170 .or. -
           BYNUM 177 .or. BYNUM 199 .or. BYNUM 206 .or. BYNUM 220 .or. -
           BYNUM 237 .or. BYNUM 247 .or. BYNUM 264 .or. BYNUM 284 .or. -
           BYNUM 303 .or. BYNUM 320 .or. BYNUM 335 .or. BYNUM 357 END
        """
        ]
    )

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(
            selectionstring,
            self.ref_selectionstring,
            err_msg="Charmm selection was not written correctly",
        )


class TestSelectionWriter_PyMOL(_SelectionWriter):

    writer = MDAnalysis.selections.pymol.SelectionWriter
    filename = "CA.pml"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one(
        [
            """# MDAnalysis PyMol selection\n select CA_selection, \\
           index 5 | index 22 | index 46 | index 65 | index 84 | index 103 | \\
           index 122 | index 129 | index 141 | index 153 | index 160 | index 170 | \\
           index 177 | index 199 | index 206 | index 220 | index 237 | index 247 | \\
           index 264 | index 284 | index 303 | index 320 | index 335 | index 357
        """
        ]
    )

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(
            selectionstring,
            self.ref_selectionstring,
            err_msg="PyMOL selection was not written correctly",
        )


class TestSelectionWriter_VMD(_SelectionWriter):

    writer = MDAnalysis.selections.vmd.SelectionWriter
    filename = "CA.vmd"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one(
        [
            """# MDAnalysis VMD selection atomselect macro CA_selection {index 4 21 45 64 83 102 121 128 \\
           140 152 159 169 176 198 205 219 \\
           236 246 263 283 302 319 334 356 }
        """
        ]
    )

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(
            selectionstring,
            self.ref_selectionstring,
            err_msg="PyMOL selection was not written correctly",
        )


def spt2array(line):
    """Get name of and convert Jmol SPT definition to integer array"""
    match = re.search(r"\@~(\w+) \(\{([\d\s]*)\}\)", line)
    return match.group(1), np.array(match.group(2).split(), dtype=int)


class TestSelectionWriter_Jmol(_SelectionWriter):

    writer = MDAnalysis.selections.jmol.SelectionWriter
    filename = "CA.spt"
    ref_name, ref_indices = spt2array(
        (
            "@~ca ({4 21 45 64 83 102 121 128 140 152 159 169 176 198 205 219 236"
            " 246 263 283 302 319 334 356});"
        )
    )

    def _assert_selectionstring(self, namedfile):
        header, indices = spt2array(namedfile.readline())
        assert_equal(
            header, self.ref_name, err_msg="SPT file has wrong selection name"
        )
        assert_array_equal(
            indices,
            self.ref_indices,
            err_msg="SPT indices were not written correctly",
        )


class TestSelections:
    @patch.object(selections, "_SELECTION_WRITERS", {"FOO": "BAR"})
    def test_get_writer_valid(self):
        writer = selections.get_writer(None, "FOO")
        assert writer == "BAR"



from io import StringIO
from collections import defaultdict

from numpy.testing import (
    assert_equal,
    assert_array_equal,
)
import pytest

import MDAnalysis
from MDAnalysis.analysis.hydrogenbonds.wbridge_analysis import (
    WaterBridgeAnalysis,
)


class TestWaterBridgeAnalysis(object):
    @staticmethod
    @pytest.fixture(scope="class")
    def universe_empty():
        """A universe with no hydrogen bonds"""
        grofile = """Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   3.000   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_DA():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        donor"""
        grofile = """Test gro file
3
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    4ALA      O    3   0.300   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_DA_PBC():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        donor but in a PBC condition"""
        grofile = """Test gro file
3
    1ALA      N    1   0.800   0.000   0.000
    1ALA      H    2   0.900   0.000   0.000
    4ALA      O    3   0.100   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AD():
        """A universe with one hydrogen bond donor bonding to a hydrogen bond
        acceptor"""
        grofile = """Test gro file
3
    1ALA      O    1   0.000   0.000   0.000
    4ALA      H    2   0.200   0.000   0.000
    4ALA      N    3   0.300   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_loop():
        """A universe with one hydrogen bond acceptor bonding to a water which
        bonds back to the first hydrogen bond  acceptor and thus form a loop"""
        grofile = """Test gro file
5
    1ALA      O    1   0.000   0.001   0.000
    2SOL     OW    2   0.300   0.001   0.000
    2SOL    HW1    3   0.200   0.002   0.000
    2SOL    HW2    4   0.200   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_DWA():
        """A universe with one hydrogen bond donor bonding to a hydrogen bond
        acceptor through a water"""
        grofile = """Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_DWD():
        """A universe with one hydrogen bond donor bonding to a hydrogen bond
        donor through a water"""
        grofile = """Test gro file
5
    1ALA      N    1   0.000   0.000   0.000
    1ALA      H    2   0.100   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWA():
        """A universe with two hydrogen bond acceptor are joined by a water"""
        grofile = """Test gro file
5
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWD():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen
        bond donor through a water"""
        grofile = """Test gro file
5
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    4ALA      H    4   0.500   0.000   0.000
    4ALA      N    5   0.600   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWWA():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through two waters"""
        grofile = """Test gro file
7
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4ALA      O    7   0.900   0.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWWWA():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through three waters"""
        grofile = """Test gro file
9
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4SOL     OW    7   0.900   0.000   0.000
    4SOL    HW1    8   1.000   0.000   0.000
    5ALA      O    9   1.200   0.000   0.000
  10.0   10.0   10.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWWWWA():
        """A universe with one hydrogen bond acceptor bonding to a hydrogen bond
        acceptor through three waters"""
        grofile = """Test gro file
11
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    4SOL     OW    7   0.900   0.000   0.000
    4SOL    HW1    8   1.000   0.000   0.000
    5SOL     OW    9   1.200   0.000   0.000
    5SOL    HW1   10   1.300   0.000   0.000
    6ALA      O   11   1.400   0.000   0.000
  10.0   10.0   10.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_branch():
        """A universe with one hydrogen bond acceptor bonding to two hydrogen
        bond acceptor in selection 2"""
        grofile = """Test gro file
9
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    3SOL     OW    5   0.600   0.000   0.000
    3SOL    HW1    6   0.700   0.000   0.000
    3SOL    HW2    7   0.600   0.100   0.000
    4ALA      O    8   0.900   0.000   0.000
    5ALA      O    9   0.600   0.300   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def universe_AWA_AWWA():
        """A universe with one hydrogen bond acceptors are bonded through one or
        two water"""
        grofile = """Test gro file
12
    1ALA      O    1   0.000   0.000   0.000
    2SOL     OW    2   0.300   0.000   0.000
    2SOL    HW1    3   0.200   0.000   0.000
    2SOL    HW2    4   0.400   0.000   0.000
    4ALA      O    5   0.600   0.000   0.000
    5ALA      O    6   0.000   1.000   0.000
    6SOL     OW    7   0.300   1.000   0.000
    6SOL    HW1    8   0.200   1.000   0.000
    6SOL    HW2    9   0.400   1.000   0.000
    7SOL     OW   10   0.600   1.000   0.000
    7SOL    HW1   11   0.700   1.000   0.000
    8ALA      O   12   0.900   1.000   0.000
  1.0   1.0   1.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def wb_multiframe():
        """A water bridge object with multipley frames"""
        grofile = """Test gro file
 13
    1ALA      O    1   0.000   0.000   0.000
    1ALA      H    2   0.000   0.000   0.000
    2SOL     OW    3   0.300   0.000   0.000
    2SOL    HW1    4   0.200   0.000   0.000
    2SOL    HW2    5   0.400   0.000   0.000
    3SOL     OW    6   0.600   0.000   0.000
    3SOL    HW1    7   0.700   0.000   0.000
    4SOL     OW    8   0.900   0.000   0.000
    4SOL    HW1    9   1.000   0.000   0.000
    5SOL     OW   10   1.200   0.000   0.000
    5SOL    HW1   11   1.300   0.000   0.000
    6ALA      H   12   1.400   0.000   0.000
    6ALA      O   13   1.400   0.000   0.000
   10.0   10.0   10.0"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        wb = WaterBridgeAnalysis(
            u, "protein and (resid 1)", "protein and (resid 4)", order=4
        )
        # Build an dummy WaterBridgeAnalysis object for testing
        wb.results.network = []
        wb.results.network.append({(1, 0, 12, None, 2.0, 180.0): None})
        wb.results.network.append({(0, None, 12, 13, 2.0, 180.0): None})
        wb.results.network.append(
            {(1, 0, 3, None, 2.0, 180.0): {(4, 2, 12, None, 2.0, 180.0): None}}
        )
        wb.results.network.append(
            {
                (0, None, 3, 2, 2.0, 180.0): {
                    (4, 2, 5, None, 2.0, 180.0): {
                        (5, None, 11, 12, 2.0, 180.0): None
                    }
                }
            }
        )
        wb.timesteps = range(len(wb.results.network))
        return wb

    def test_nodata(self, universe_DA):
        """Test if the funtions can run when there is no data.
        This is achieved by not runing the run() first."""
        wb = WaterBridgeAnalysis(
            universe_DA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=0,
        )
        wb.generate_table()
        assert_equal(wb.timesteps_by_type(), None)
        assert_equal(wb.count_by_time(), None)
        assert_equal(wb.count_by_type(), None)

    def test_selection_type_error(self, universe_DA):
        """Test the case when the wrong selection1_type is given"""
        try:
            wb = WaterBridgeAnalysis(
                universe_DA,
                "protein and (resid 1)",
                "protein and (resid 4)",
                order=0,
                selection1_type="aaa",
            )
        except ValueError:
            pass
        else:
            raise pytest.fail("selection_type aaa should rasie error")

    def test_distance_type_error(self, universe_DA):
        """Test the case when the wrong selection1_type is given"""
        with pytest.raises(
            ValueError,
            match="Only 'hydrogen' and 'heavy' are allowed for option `distance_type'",
        ):
            WaterBridgeAnalysis(
                universe_DA,
                "protein and (resid 1)",
                "protein and (resid 4)",
                order=0,
                selection1_type="aaa",
                distance_type="aaa",
            )

    def test_selection2_type_error(self, universe_DA):
        """Test the case when the wrong selection1_type is given"""
        with pytest.raises(
            ValueError, match="`selection2_type` is not a keyword argument."
        ):
            WaterBridgeAnalysis(
                universe_DA,
                "protein and (resid 1)",
                "protein and (resid 4)",
                order=0,
                selection1_type="aaa",
                selection2_type="aaa",
            )

    def test_empty_selection(self, universe_DA):
        """Test the case when selection yields empty result"""
        wb = WaterBridgeAnalysis(
            universe_DA,
            "protein and (resid 9)",
            "protein and (resid 10)",
            order=0,
        )
        wb.run()
        assert wb.results.network == [{}]

    def test_loop(self, universe_loop):
        """Test if loop can be handled correctly"""
        wb = WaterBridgeAnalysis(
            universe_loop,
            "protein and (resid 1)",
            "protein and (resid 1 or resid 4)",
        )
        wb.run()
        assert_equal(len(wb.results.network[0].keys()), 2)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_donor_accepter(self, universe_DA, distance_type):
        """Test zeroth order donor to acceptor hydrogen bonding"""
        wb = WaterBridgeAnalysis(
            universe_DA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=0,
            update_selection=True,
            debug=True,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_donor_accepter_pbc(self, universe_DA_PBC, distance_type):
        """Test zeroth order donor to acceptor hydrogen bonding in PBC conditions"""
        wb = WaterBridgeAnalysis(
            universe_DA_PBC,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=0,
            pbc=True,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_accepter_donor(self, universe_AD, distance_type):
        """Test zeroth order acceptor to donor hydrogen bonding"""
        wb = WaterBridgeAnalysis(
            universe_AD,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=0,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 1, 2))

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_water_accepter(self, universe_AWA, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form
        water bridge with hydrogen bond acceptor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_AWA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        assert_equal(second[list(second.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_donor_water_accepter(self, universe_DWA, distance_type):
        """Test case where the hydrogen bond donor from selection 1 form
        water bridge with hydrogen bond acceptor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_DWA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 2, 4, None))
        assert_equal(second[list(second.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_water_donor(self, universe_AWD, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form
        water bridge with hydrogen bond donor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_AWD,
            "protein and (resid 1)",
            "protein and (resid 4)",
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (1, None, 3, 4))
        assert_equal(second[list(second.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_donor_water_donor(self, universe_DWD, distance_type):
        """Test case where the hydrogen bond donor from selection 1 form
        water bridge with hydrogen bond donor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_DWD,
            "protein and (resid 1)",
            "protein and (resid 4)",
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (1, 0, 2, None))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (2, None, 3, 4))
        assert_equal(second[list(second.keys())[0]], None)

    def test_empty(self, universe_empty):
        """Test case where no water bridge exists"""
        wb = WaterBridgeAnalysis(universe_empty, "protein", "protein")
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

    def test_same_selection(self, universe_DWA):
        """
        This test tests that if the selection 1 and selection 2 are both protein.
        However, the protein only forms one hydrogen bond with the water.
        This entry won't be included.
        """
        wb = WaterBridgeAnalysis(
            universe_DWA, "protein and resid 1", "protein and resid 1"
        )
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_2water_accepter(self, universe_AWWA, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form second order
        water bridge with hydrogen bond acceptor from selection 2"""
        # test first order
        wb = WaterBridgeAnalysis(
            universe_AWWA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))
        # test second order
        wb = WaterBridgeAnalysis(
            universe_AWWA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=2,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        assert_equal(third[list(third.keys())[0]], None)
        # test third order
        wb = WaterBridgeAnalysis(
            universe_AWWA,
            "protein and (resid 1)",
            "protein and (resid 4)",
            order=3,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        assert_equal(third[list(third.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_3water_accepter(self, universe_AWWWA, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form third order
        water bridge with hydrogen bond acceptor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_AWWWA,
            "protein and (resid 1)",
            "protein and (resid 5)",
            order=2,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

        wb = WaterBridgeAnalysis(
            universe_AWWWA,
            "protein and (resid 1)",
            "protein and (resid 5)",
            order=3,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        assert_equal(fourth[list(fourth.keys())[0]], None)

        wb = WaterBridgeAnalysis(
            universe_AWWWA,
            "protein and (resid 1)",
            "protein and (resid 5)",
            order=4,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        assert_equal(fourth[list(fourth.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_4water_accepter(self, universe_AWWWWA, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form fourth order
        water bridge with hydrogen bond acceptor from selection 2"""
        wb = WaterBridgeAnalysis(
            universe_AWWWWA,
            "protein and (resid 1)",
            "protein and (resid 6)",
            order=3,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        assert_equal(wb.results.network[0], defaultdict(dict))

        wb = WaterBridgeAnalysis(
            universe_AWWWWA,
            "protein and (resid 1)",
            "protein and (resid 6)",
            order=4,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        fifth = fourth[list(fourth.keys())[0]]
        assert_equal(list(fifth.keys())[0][:4], (9, 8, 10, None))
        assert_equal(fifth[list(fifth.keys())[0]], None)

        wb = WaterBridgeAnalysis(
            universe_AWWWWA,
            "protein and (resid 1)",
            "protein and (resid 6)",
            order=5,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(list(third.keys())[0][:4], (5, 4, 6, None))
        fourth = third[list(third.keys())[0]]
        assert_equal(list(fourth.keys())[0][:4], (7, 6, 8, None))
        fifth = fourth[list(fourth.keys())[0]]
        assert_equal(list(fifth.keys())[0][:4], (9, 8, 10, None))
        assert_equal(fifth[list(fifth.keys())[0]], None)

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_22water_accepter(self, universe_branch, distance_type):
        """Test case where the hydrogen bond acceptor from selection 1 form a second order
        water bridge with hydrogen bond acceptor from selection 2
        and the last water is linked to two residues in selection 2"""
        wb = WaterBridgeAnalysis(
            universe_branch,
            "protein and (resid 1)",
            "protein and (resid 4 or resid 5)",
            order=2,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        third = second[list(second.keys())[0]]
        assert_equal(
            [(5, 4, 7, None), (6, 4, 8, None)],
            sorted([key[:4] for key in list(third.keys())]),
        )

    def test_timeseries_wba(self, universe_branch):
        """Test if the time series data is correctly generated in water bridge analysis format"""
        wb = WaterBridgeAnalysis(
            universe_branch,
            "protein and (resid 1)",
            "protein and (resid 4 or resid 5)",
            order=2,
        )
        wb.output_format = "sele1_sele2"
        wb.run(verbose=False)
        timeseries = sorted(wb.results.timeseries[0])

        assert_equal(
            timeseries[0][:4], (0, 2, ("ALA", 1, "O"), ("SOL", 2, "HW1"))
        )
        assert_equal(
            timeseries[1][:4], (3, 4, ("SOL", 2, "HW2"), ("SOL", 3, "OW"))
        )
        assert_equal(
            timeseries[2][:4], (5, 7, ("SOL", 3, "HW1"), ("ALA", 4, "O"))
        )
        assert_equal(
            timeseries[3][:4], (6, 8, ("SOL", 3, "HW2"), ("ALA", 5, "O"))
        )

    def test_timeseries_hba(self, universe_branch):
        """Test if the time series data is correctly generated in hydrogen bond analysis format"""
        wb = WaterBridgeAnalysis(
            universe_branch,
            "protein and (resid 1)",
            "protein and (resid 4 or resid 5)",
            order=2,
        )
        wb.output_format = "donor_acceptor"
        wb.run(verbose=False)
        timeseries = sorted(wb.results.timeseries[0])

        assert_equal(
            timeseries[0][:4], (2, 0, ("SOL", 2, "HW1"), ("ALA", 1, "O"))
        )
        assert_equal(
            timeseries[1][:4], (3, 4, ("SOL", 2, "HW2"), ("SOL", 3, "OW"))
        )
        assert_equal(
            timeseries[2][:4], (5, 7, ("SOL", 3, "HW1"), ("ALA", 4, "O"))
        )
        assert_equal(
            timeseries[3][:4], (6, 8, ("SOL", 3, "HW2"), ("ALA", 5, "O"))
        )

    @pytest.mark.parametrize("distance_type", ["hydrogen", "heavy"])
    def test_acceptor_12water_accepter(self, universe_AWA_AWWA, distance_type):
        """Test of independent first order and second can be recognised correctely"""
        wb = WaterBridgeAnalysis(
            universe_AWA_AWWA,
            "protein and (resid 1 or resid 5)",
            "protein and (resid 4 or resid 8)",
            order=1,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(list(network.keys())[0][:4], (0, None, 2, 1))
        second = network[list(network.keys())[0]]
        assert_equal(list(second.keys())[0][:4], (3, 1, 4, None))
        assert_equal(second[list(second.keys())[0]], None)
        network = wb.results.network[0]
        wb = WaterBridgeAnalysis(
            universe_AWA_AWWA,
            "protein and (resid 1 or resid 5)",
            "protein and (resid 4 or resid 8)",
            order=2,
            distance_type=distance_type,
        )
        wb.run(verbose=False)
        network = wb.results.network[0]
        assert_equal(
            [(0, None, 2, 1), (5, None, 7, 6)],
            sorted([key[:4] for key in list(network.keys())]),
        )

    def test_count_by_type_single_link(self, universe_DWA):
        """
        This test tests the simplest water bridge to see if count_by_type() works.
        """
        wb = WaterBridgeAnalysis(
            universe_DWA, "protein and (resid 1)", "protein and (resid 4)"
        )
        wb.run(verbose=False)
        assert_equal(
            wb.count_by_type(), [(1, 4, "ALA", 1, "H", "ALA", 4, "O", 1.0)]
        )

    def test_count_by_type_multiple_link(self, universe_AWA_AWWA):
        """
        This test tests if count_by_type() can give the correct result for more than 1 links.
        """
        wb = WaterBridgeAnalysis(
            universe_AWA_AWWA,
            "protein and (resid 1 or resid 5)",
            "protein and (resid 4 or resid 8)",
            order=2,
        )
        wb.run(verbose=False)
        assert_equal(
            sorted(wb.count_by_type()),
            [
                [0, 4, "ALA", 1, "O", "ALA", 4, "O", 1.0],
                [5, 11, "ALA", 5, "O", "ALA", 8, "O", 1.0],
            ],
        )

    def test_count_by_type_multiple_frame(self, wb_multiframe):
        """
        This test tests if count_by_type() works in multiply situations.
        :return:
        """
        result = [
            [0, 11, "ALA", 1, "O", "ALA", 6, "H", 0.25],
            [0, 12, "ALA", 1, "O", "ALA", 6, "O", 0.25],
            [1, 12, "ALA", 1, "H", "ALA", 6, "O", 0.5],
        ]

        assert_equal(sorted(wb_multiframe.count_by_type()), result)

    def test_count_by_type_filter(self, wb_multiframe):
        """
        This test tests if modifying analysis_func
        allows some results to be filtered out in count_by_type().
        :return:
        """

        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = (
                current[0]
            )
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = (
                current[-1]
            )
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (
                sele1.resname,
                sele1.resid,
                sele1.name,
            )
            (s2_resname, s2_resid, s2_name) = (
                sele2.resname,
                sele2.resid,
                sele2.name,
            )

            key = (
                sele1_index,
                sele2_index,
                s1_resname,
                s1_resid,
                s1_name,
                s2_resname,
                s2_resid,
                s2_name,
            )
            if s2_name == "H":
                output[key] += 1

        result = [((0, 11, "ALA", 1, "O", "ALA", 6, "H"), 0.25)]
        assert_equal(
            sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result
        )

    def test_count_by_type_merge(self, wb_multiframe):
        """
        This test tests if modifying analysis_func
        allows some same residue to be merged in count_by_type().
        """

        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = (
                current[0]
            )
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = (
                current[-1]
            )
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (
                sele1.resname,
                sele1.resid,
                sele1.name,
            )
            (s2_resname, s2_resid, s2_name) = (
                sele2.resname,
                sele2.resid,
                sele2.name,
            )

            key = (s1_resname, s1_resid, s2_resname, s2_resid)
            output[key] = 1

        result = [(("ALA", 1, "ALA", 6), 1.0)]
        assert_equal(
            sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result
        )

    def test_count_by_type_order(self, wb_multiframe):
        """
        This test tests if modifying analysis_func
        allows the order of water bridge to be separated in count_by_type().
        :return:
        """

        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = (
                current[0]
            )
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = (
                current[-1]
            )
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (
                sele1.resname,
                sele1.resid,
                sele1.name,
            )
            (s2_resname, s2_resid, s2_name) = (
                sele2.resname,
                sele2.resid,
                sele2.name,
            )
            key = (
                s1_resname,
                s1_resid,
                s2_resname,
                s2_resid,
                len(current) - 1,
            )
            output[key] = 1

        result = [
            (("ALA", 1, "ALA", 6, 0), 0.5),
            (("ALA", 1, "ALA", 6, 1), 0.25),
            (("ALA", 1, "ALA", 6, 2), 0.25),
        ]
        assert_equal(
            sorted(wb_multiframe.count_by_type(analysis_func=analysis)), result
        )

    def test_count_by_time(self, wb_multiframe):
        """
        This test tests if count_by_times() works.
        :return:
        """
        assert_equal(
            wb_multiframe.count_by_time(), [(0, 1), (1, 1), (2, 1), (3, 1)]
        )

    def test_count_by_time_weight(self, universe_AWA_AWWA):
        """
        This test tests if modyfing the analysis_func allows the weight to be changed
        in count_by_type().
        :return:
        """
        wb = WaterBridgeAnalysis(
            universe_AWA_AWWA,
            "protein and (resid 1 or resid 5)",
            "protein and (resid 4 or resid 8)",
            order=2,
        )
        wb.run(verbose=False)

        def analysis(current, output, u):
            sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle = (
                current[0]
            )
            atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle = (
                current[-1]
            )
            sele1 = u.atoms[sele1_index]
            sele2 = u.atoms[sele2_index]
            (s1_resname, s1_resid, s1_name) = (
                sele1.resname,
                sele1.resid,
                sele1.name,
            )
            (s2_resname, s2_resid, s2_name) = (
                sele2.resname,
                sele2.resid,
                sele2.name,
            )
            key = (s1_resname, s1_resid, s2_resname, s2_resid)
            output[key] += len(current) - 1

        assert_equal(
            wb.count_by_time(analysis_func=analysis),
            [
                (0, 3),
            ],
        )

    def test_count_by_time_empty(self, universe_AWA_AWWA):
        """
        See if count_by_time() can handle zero well.
        :return:
        """
        wb = WaterBridgeAnalysis(
            universe_AWA_AWWA,
            "protein and (resid 1 or resid 5)",
            "protein and (resid 4 or resid 8)",
            order=2,
        )
        wb.run(verbose=False)

        def analysis(current, output, u):
            pass

        assert_equal(
            wb.count_by_time(analysis_func=analysis),
            [
                (0, 0),
            ],
        )

    def test_generate_table_hba(self, wb_multiframe):
        """Test generate table using hydrogen bond analysis format"""
        table = wb_multiframe.generate_table(output_format="donor_acceptor")
        assert_array_equal(
            sorted(table.donor_resid),
            [1, 1, 2, 2, 2, 6, 6],
        )

    def test_generate_table_s1s2(self, wb_multiframe):
        """Test generate table using hydrogen bond analysis format"""
        table = wb_multiframe.generate_table(output_format="sele1_sele2")
        assert_array_equal(
            sorted(table.sele1_resid),
            [1, 1, 1, 1, 2, 2, 3],
        )

    def test_timesteps_by_type(self, wb_multiframe):
        """Test the timesteps_by_type function"""

        timesteps = sorted(wb_multiframe.timesteps_by_type())
        assert_array_equal(
            timesteps[3], [1, 12, "ALA", 1, "H", "ALA", 6, "O", 0, 2]
        )

    def test_duplicate_water(self):
        """A case #3119 where
        Acceptor···H−O···H-Donor
                     |
                     H···O-H
        will be recognised as 3rd order water bridge.
        """
        grofile = """Test gro file
    7
    1LEU      O    1   1.876   0.810   1.354
  117SOL    HW1    2   1.853   0.831   1.162
  117SOL     OW    3   1.877   0.890   1.081
  117SOL    HW2    4   1.908   0.828   1.007
  135SOL     OW    5   1.924   0.713   0.845
    1LEU      H    6   1.997   0.991   1.194
    1LEU      N    7   2.041   1.030   1.274
   2.22092   2.22092   2.22092"""
        u = MDAnalysis.Universe(StringIO(grofile), format="gro")
        wb = WaterBridgeAnalysis(
            u, "resname LEU and name O", "resname LEU and name N H", order=4
        )
        wb.run()
        assert len(wb.results.timeseries[0]) == 2

    def test_warn_results_deprecated(self, universe_DA):
        wb = WaterBridgeAnalysis(
            universe_DA,
            "protein and (resid 9)",
            "protein and (resid 10)",
            order=0,
        )
        wb.run()

        wmsg = "The `network` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(wb.network, wb.results.network)

        wmsg = "The `timeseries` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(wb.timeseries, wb.results.timeseries)



try:
    from MDAnalysis.coordinates.DCD import DCDReader
    from MDAnalysisTests.datafiles import DCD
except ImportError:
    pass

try:
    from MDAnalysis.coordinates.XTC import XTCReader
    from MDAnalysisTests.datafiles import XTC
except ImportError:
    pass

try:
    from MDAnalysis.coordinates.TRR import TRRReader
    from MDAnalysisTests.datafiles import TRR
except ImportError:
    pass

try:
    from MDAnalysis.coordinates.TRJ import NCDFReader
    from MDAnalysisTests.datafiles import NCDF
except ImportError:
    pass

try:
    from MDAnalysis.coordinates.TRC import TRCReader
    from MDAnalysisTests.datafiles import TRC_TRAJ_SOLV
except ImportError:
    pass

traj_dict = {
    "XTC": [XTC, XTCReader],
    "TRR": [TRR, TRRReader],
    "DCD": [DCD, DCDReader],
    "NCDF": [NCDF, NCDFReader],
    "TRC": [TRC_TRAJ_SOLV, TRCReader],
}


class TrajReaderCreation(object):
    """Benchmarks for trajectory file format reading."""

    params = ["XTC", "TRR", "DCD", "NCDF", "TRC"]
    param_names = ["traj_format"]

    def setup(self, traj_format):
        self.traj_dict = traj_dict
        self.traj_file, self.traj_reader = self.traj_dict[traj_format]

    def time_reads(self, traj_format):
        """Simple benchmark for reading traj file formats
        from our standard test files.
        """
        self.traj_reader(self.traj_file)


class TrajReaderIteration(object):
    """Benchmarks for trajectory file format striding."""

    params = ["XTC", "TRR", "DCD", "NCDF", "TRC"]
    param_names = ["traj_format"]

    def setup(self, traj_format):
        self.traj_dict = traj_dict
        self.traj_file, self.traj_reader = self.traj_dict[traj_format]
        self.reader_object = self.traj_reader(self.traj_file)

    def time_strides(self, traj_format):
        """Benchmark striding over full trajectory
        test files for each format.
        """
        for ts in self.reader_object:
            pass



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from pytest import approx
import MDAnalysis as mda

from numpy.testing import assert_equal, assert_allclose
import numpy as np
from MDAnalysis.core.topologyattrs import Angles, Atomtypes, Atomnames
from MDAnalysis.exceptions import NoDataError
from MDAnalysis.guesser.default_guesser import DefaultGuesser
from MDAnalysis.core.topology import Topology
from MDAnalysisTests import make_Universe
from MDAnalysisTests.core.test_fragments import make_starshape
import MDAnalysis.tests.datafiles as datafiles
from MDAnalysisTests.util import import_not_available

try:
    from rdkit import Chem
    from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges
except ImportError:
    pass

requires_rdkit = pytest.mark.skipif(
    import_not_available("rdkit"), reason="requires RDKit"
)


@pytest.fixture
def default_guesser():
    return DefaultGuesser(None)


class TestGuessMasses(object):
    def test_guess_masses_from_universe(self):
        topology = Topology(3, attrs=[Atomtypes(["C", "C", "H"])])
        u = mda.Universe(topology)

        assert isinstance(u.atoms.masses, np.ndarray)
        assert_allclose(
            u.atoms.masses, np.array([12.011, 12.011, 1.008]), atol=0
        )

    def test_guess_masses_from_guesser_object(self, default_guesser):
        elements = ["H", "Ca", "Am"]
        values = np.array([1.008, 40.08000, 243.0])
        assert_allclose(default_guesser.guess_masses(elements), values, atol=0)

    def test_guess_masses_warn(self):
        topology = Topology(2, attrs=[Atomtypes(["X", "Z"])])
        msg = "Unknown masses are set to 0.0 for current version, "
        "this will be depracated in version 3.0.0 and replaced by"
        " Masse's no_value_label (np.nan)"
        with pytest.warns(PendingDeprecationWarning, match=msg):
            u = mda.Universe(topology, to_guess=["masses"])
            assert_allclose(u.atoms.masses, np.array([0.0, 0.0]), atol=0)

    @pytest.mark.parametrize(
        "element, value",
        (
            ("H", 1.008),
            ("XYZ", 0.0),
        ),
    )
    def test_get_atom_mass(self, element, value, default_guesser):
        default_guesser.get_atom_mass(element) == approx(value)

    def test_guess_atom_mass(self, default_guesser):
        assert default_guesser.guess_atom_mass("1H") == approx(1.008)

    def test_guess_masses_with_no_reference_elements(self):
        u = mda.Universe.empty(3)
        with pytest.raises(
            NoDataError, match=("there is no reference attributes ")
        ):
            u.guess_TopologyAttrs("default", ["masses"])


class TestGuessTypes(object):

    def test_guess_types(self):
        topology = Topology(2, attrs=[Atomnames(["MG2+", "C12"])])
        u = mda.Universe(topology, to_guess=["types"])
        assert isinstance(u.atoms.types, np.ndarray)
        assert_equal(u.atoms.types, np.array(["MG", "C"], dtype=object))

    def test_guess_atom_element(self, default_guesser):
        assert default_guesser.guess_atom_element("MG2+") == "MG"

    def test_guess_atom_element_empty(self, default_guesser):
        assert default_guesser.guess_atom_element("") == ""

    def test_guess_atom_element_singledigit(self, default_guesser):
        assert default_guesser.guess_atom_element("1") == "1"

    def test_guess_atom_element_1H(self, default_guesser):
        assert default_guesser.guess_atom_element("1H") == "H"
        assert default_guesser.guess_atom_element("2H") == "H"

    def test_partial_guess_elements(self, default_guesser):
        names = np.array(["BR123", "Hk", "C12"], dtype=object)
        elements = np.array(["BR", "C"], dtype=object)
        guessed_elements = default_guesser.guess_types(
            atom_types=names, indices_to_guess=[True, False, True]
        )
        assert_equal(elements, guessed_elements)

    def test_guess_elements_from_no_data(self):
        top = Topology(5)
        msg = (
            "there is no reference attributes in this "
            "universe to guess types from"
        )
        with pytest.warns(UserWarning, match=msg):
            mda.Universe(top, to_guess=["types"])

    @pytest.mark.parametrize(
        "name, element",
        (
            ("AO5*", "O"),
            ("F-", "F"),
            ("HB1", "H"),
            ("OC2", "O"),
            ("1he2", "H"),
            ("3hg2", "H"),
            ("OH-", "O"),
            ("HO", "H"),
            ("he", "H"),
            ("zn", "ZN"),
            ("Ca2+", "CA"),
            ("CA", "C"),
        ),
    )
    def test_guess_element_from_name(self, name, element, default_guesser):
        assert default_guesser.guess_atom_element(name) == element


def test_guess_charge(default_guesser):
    # this always returns 0.0
    assert default_guesser.guess_atom_charge("this") == approx(0.0)


def test_guess_bonds_Error():
    u = make_Universe(trajectory=True)
    msg = "This Universe does not contain name information"
    with pytest.raises(NoDataError, match=msg):
        u.guess_TopologyAttrs(to_guess=["bonds"])


def test_guess_bond_vdw_error():
    u = mda.Universe(datafiles.PDB)
    with pytest.raises(ValueError, match="vdw radii for types: DUMMY"):
        DefaultGuesser(u).guess_bonds(u.atoms)


def test_guess_bond_coord_error(default_guesser):
    msg = "atoms' and 'coord' must be the same length"
    with pytest.raises(ValueError, match=msg):
        default_guesser.guess_bonds(["N", "O", "C"], [[1, 2, 3]])


def test_guess_angles_with_no_bonds():
    "Test guessing angles for atoms with no bonds"
    " information without adding bonds to universe "
    u = mda.Universe(datafiles.two_water_gro)
    u.guess_TopologyAttrs(to_guess=["angles"])
    assert hasattr(u, "angles")
    assert not hasattr(u, "bonds")


def test_guess_impropers(default_guesser):
    u = make_starshape()

    ag = u.atoms[:5]
    guessed_angles = default_guesser.guess_angles(ag.bonds)
    u.add_TopologyAttr(Angles(guessed_angles))

    vals = default_guesser.guess_improper_dihedrals(ag.angles)
    assert_equal(len(vals), 12)


def test_guess_dihedrals_with_no_angles():
    "Test guessing dihedrals for atoms with no angles"
    "information without adding bonds or angles to universe"
    u = mda.Universe(datafiles.two_water_gro)
    u.guess_TopologyAttrs(to_guess=["dihedrals"])
    assert hasattr(u, "dihedrals")
    assert not hasattr(u, "angles")
    assert not hasattr(u, "bonds")


def test_guess_impropers_with_angles():
    "Test guessing impropers for atoms with angles"
    "and bonds information "
    u = mda.Universe(
        datafiles.two_water_gro, to_guess=["bonds", "angles", "impropers"]
    )
    u.guess_TopologyAttrs(to_guess=["impropers"])
    assert hasattr(u, "impropers")
    assert hasattr(u, "angles")
    assert hasattr(u, "bonds")


def test_guess_impropers_with_no_angles():
    "Test guessing impropers for atoms with no angles"
    "information without adding bonds or angles to universe"
    u = mda.Universe(datafiles.two_water_gro)
    u.guess_TopologyAttrs(to_guess=["impropers"])
    assert hasattr(u, "impropers")
    assert not hasattr(u, "angles")
    assert not hasattr(u, "bonds")


def bond_sort(arr):
    # sort from low to high, also within a tuple
    # e.g. ([5, 4], [0, 1], [0, 3]) -> ([0, 1], [0, 3], [4, 5])
    out = []
    for i, j in arr:
        if i > j:
            i, j = j, i
        out.append((i, j))
    return sorted(out)


def test_guess_bonds_water():
    u = mda.Universe(datafiles.two_water_gro)
    bonds = bond_sort(
        DefaultGuesser(None, box=u.dimensions).guess_bonds(
            u.atoms, u.atoms.positions
        )
    )
    assert_equal(bonds, ((0, 1), (0, 2), (3, 4), (3, 5)))


@pytest.mark.parametrize(
    "fudge_factor, n_bonds", [(0, 0), (0.55, 4), (200, 6)]
)
def test_guess_bonds_water_fudge_factor_passed(fudge_factor, n_bonds):
    u = mda.Universe(
        datafiles.two_water_gro,
        fudge_factor=fudge_factor,
        to_guess=("types", "bonds"),
    )
    assert len(u.atoms.bonds) == n_bonds


def test_guess_bonds_adk():
    u = mda.Universe(datafiles.PSF, datafiles.DCD)
    u.guess_TopologyAttrs(force_guess=["types"])
    guesser = DefaultGuesser(None)
    bonds = bond_sort(guesser.guess_bonds(u.atoms, u.atoms.positions))
    assert_equal(np.sort(u.bonds.indices, axis=0), np.sort(bonds, axis=0))


def test_guess_bonds_peptide():
    u = mda.Universe(datafiles.PSF_NAMD, datafiles.PDB_NAMD)
    u.guess_TopologyAttrs(force_guess=["types"])
    guesser = DefaultGuesser(None)
    bonds = bond_sort(guesser.guess_bonds(u.atoms, u.atoms.positions))
    assert_equal(np.sort(u.bonds.indices, axis=0), np.sort(bonds, axis=0))


@pytest.mark.parametrize(
    "smi",
    [
        "c1ccccc1",
        "C1=CC=CC=C1",
        "CCO",
        "c1ccccc1Cc1ccccc1",
        "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
    ],
)
@requires_rdkit
def test_guess_aromaticities(smi):
    mol = Chem.MolFromSmiles(smi)
    mol = Chem.AddHs(mol)
    expected = np.array([atom.GetIsAromatic() for atom in mol.GetAtoms()])
    u = mda.Universe(mol)
    guesser = DefaultGuesser(None)
    values = guesser.guess_aromaticities(u.atoms)
    u.guess_TopologyAttrs(to_guess=["aromaticities"])
    assert_equal(values, expected)
    assert_equal(u.atoms.aromaticities, expected)


@pytest.mark.parametrize(
    "smi",
    [
        "c1ccccc1",
        "C1=CC=CC=C1",
        "CCO",
        "c1ccccc1Cc1ccccc1",
        "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
    ],
)
@requires_rdkit
def test_guess_gasteiger_charges(smi):
    mol = Chem.MolFromSmiles(smi)
    mol = Chem.AddHs(mol)
    ComputeGasteigerCharges(mol, throwOnParamFailure=True)
    expected = np.array(
        [atom.GetDoubleProp("_GasteigerCharge") for atom in mol.GetAtoms()],
        dtype=np.float32,
    )
    u = mda.Universe(mol)
    guesser = DefaultGuesser(None)
    values = guesser.guess_gasteiger_charges(u.atoms)
    assert_equal(values, expected)


@requires_rdkit
def test_aromaticity():
    u = mda.Universe(
        datafiles.PDB_small, to_guess=["elements", "aromaticities"]
    )
    c_aromatic = u.select_atoms("resname PHE and name CD1")
    assert_equal(c_aromatic.aromaticities[0], True)



import MDAnalysis
import numpy as np

try:
    from MDAnalysisTests.datafiles import DCD, PSF
except:
    pass

try:
    from MDAnalysis.analysis import rms
except:
    pass


class SimpleRmsBench(object):
    """Benchmarks for MDAnalysis.analysis.rms.rmsd"""

    params = ([100, 500, 2000], [True, False], [False, True], [False, True])
    param_names = ["num_atoms", "use_weights", "center", "superposition"]

    def setup(self, num_atoms, use_weights, center, superposition):
        # mimic rmsd docstring example code
        self.u = MDAnalysis.Universe(PSF, DCD)
        # ag.positions is the new syntax
        # but older commit hashes will need to use
        # ag.coordinates()
        self.A = self.u.atoms.positions.copy()[:num_atoms]
        self.u.trajectory[-1]
        self.B = self.u.atoms.positions.copy()[:num_atoms]
        self.atoms = self.u.atoms[:num_atoms]
        self.weights = self.atoms.masses/np.sum(self.atoms.masses) if use_weights else None

    def time_rmsd(self, num_atoms, weights, center, superposition):
        """Benchmark rmsd function using a setup similar to
        its docstring example code along with several possible
        permutations of parameters.
        """
        rms.rmsd(
            a=self.A,
            b=self.B,
            weights=self.weights,
            center=center,
            superposition=superposition,
        )


class RmsdTrajBench(object):
    """Benchmarks for MDAnalysis.analysis.rms.RMSD"""

    # TODO: RMSD has many parameters / options,
    # some of which are apparently still considered
    # experimental -- we'll eventually want to
    # benchmark more of these

    params = (["all", "backbone"], [None, "mass"])

    param_names = ["select", "weights"]

    def setup(self, select, weights):
        self.u = MDAnalysis.Universe(PSF, DCD)
        self.RMSD_inst = rms.RMSD(
            atomgroup=self.u, reference=None, select=select, weights=weights
        )

    def time_RMSD(self, select, weights):
        """Benchmark RMSD.run() method, which parses
        over the entire trajectory.
        """
        self.RMSD_inst.run()


class RmsfTrajBench(object):
    """Benchmarks for MDAnalysis.analysis.rms.RMSF"""

    params = ([100, 500, 2000], [None, 3], [None, "mass"])

    param_names = ["n_atoms", "step", "weights"]

    def setup(self, n_atoms, step, weights):
        self.u = MDAnalysis.Universe(PSF, DCD)
        self.ag = self.u.atoms[:n_atoms]
        self.RMSF_inst = rms.RMSF(atomgroup=self.ag, weights=weights)

    def time_RMSF(self, n_atoms, step, weights):
        """Benchmark RMSF.run() method, which parses
        over the entire trajectory.
        """
        self.RMSF_inst.run(step=step)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
r"""Bond-Angle-Torsion coordinates analysis --- :mod:`MDAnalysis.analysis.bat`
===========================================================================

:Author: Soohaeng Yoo Willow and David Minh
:Year: 2020
:Copyright: Lesser GNU Public License, v2.1 or any higher version

.. versionadded:: 2.0.0

This module contains classes for interconverting between Cartesian and an
internal coordinate system, Bond-Angle-Torsion (BAT) coordinates
:footcite:p:`Chang2003`, for a given set of atoms or residues. This coordinate
system is designed to be complete, non-redundant, and minimize correlations
between degrees of freedom. Complete and non-redundant means that for N atoms
there will be 3N Cartesian coordinates and 3N BAT coordinates. Correlations are
minimized by using improper torsions, as described in :footcite:p:`Hikiri2016`.

More specifically, bond refers to the bond length, or distance between
a pair of bonded atoms. Angle refers to the bond angle, the angle between
a pair of bonds to a central atom. Torsion refers to the torsion angle.
For a set of four atoms a, b, c, and d, a torsion requires bonds between
a and b, b and c, and c and d. The torsion is the angle between a plane
containing atoms a, b, and c and another plane containing b, c, and d.
For a set of torsions that share atoms b and c, one torsion is defined as
the primary torsion. The others are defined as improper torsions, differences
between the raw torsion angle and the primary torsion. This definition reduces
the correlation between the torsion angles.

Each molecule also has six external coordinates that define its translation and
rotation in space. The three Cartesian coordinates of the first atom are the
molecule's translational degrees of freedom. Rotational degrees of freedom are
specified by the axis-angle convention. The rotation axis is a normalized vector
pointing from the first to second atom. It is described by the polar angle,
:math:`\phi`, and azimuthal angle, :math:`\theta`. :math:`\omega` is a third angle
that describes the rotation of the third atom about the axis.

This module was adapted from AlGDock :footcite:p:`Minh2020`.


See Also
--------
:class:`~MDAnalysis.analysis.dihedrals.Dihedral`
   class to calculate dihedral angles for a given set of atoms or residues
:func:`MDAnalysis.lib.distances.calc_dihedrals()`
   function to calculate dihedral angles from atom positions


Example applications
--------------------

The :class:`~MDAnalysis.analysis.bat.BAT` class defines bond-angle-torsion
coordinates based on the topology of an atom group and interconverts between
Cartesian and BAT coordinate systems.

For example, we can determine internal coordinates for residues 5-10
of adenylate kinase (AdK). The trajectory is included within the test data files::

   import MDAnalysis as mda
   from MDAnalysisTests.datafiles import PSF, DCD
   import numpy as np

   u = mda.Universe(PSF, DCD)

   # selection of atomgroups
   selected_residues = u.select_atoms("resid 5-10")

   from MDAnalysis.analysis.bat import BAT
   R = BAT(selected_residues)

   # Calculate BAT coordinates for a trajectory
   R.run()

After :meth:`R.run()<BAT.run>`, the coordinates can be accessed with
:attr:`R.results.bat<BAT.bat>`. The following code snippets assume that the
previous snippet has been executed.

Reconstruct Cartesian coordinates for the first frame::

   # Reconstruct Cartesian coordinates from BAT coordinates
   # of the first frame
   XYZ = R.Cartesian(R.results.bat[0,:])

   # The original and reconstructed Cartesian coordinates should all be close
   print(np.allclose(XYZ, selected_residues.positions, atol=1e-6))

Change a single torsion angle by :math:`\pi`::

   bat = R.results.bat[0,:]
   bat[bat.shape[0]-12] += np.pi
   XYZ = R.Cartesian(bat)

   # A good number of Cartesian coordinates should have been modified
   np.sum((XYZ - selected_residues.positions)>1E-5)

Store data to the disk and load it again::

   # BAT coordinates can be saved to disk in the numpy binary format
   R.save('test.npy')

   # The BAT coordinates in a new BAT instance can be loaded from disk
   # instead of using the run() method.
   Rnew = BAT(selected_residues, filename='test.npy')

   # The BAT coordinates before and after disk I/O should be close
   print(np.allclose(Rnew.results.bat, R.results.bat))


Analysis classes
----------------
.. autoclass:: BAT
    :members:
    :inherited-members:

    .. attribute:: results.bat

        Contains the time series of the Bond-Angle-Torsion coordinates as a
        (nframes, 3N) :class:`numpy.ndarray` array. Each row corresponds to
        a frame in the trajectory. In each column, the first six elements
        describe external degrees of freedom. The first three are the center
        of mass of the initial atom. The next three specify the  external angles
        according to the axis-angle convention: :math:`\phi`, the polar angle,
        :math:`\theta`, the azimuthal angle, and :math:`\omega`, a third angle
        that describes the rotation of the third atom about the axis. The next
        three degrees of freedom are internal degrees of freedom for the root
        atoms: :math:`r_{01}`, the distance between atoms 0 and 1,
        :math:`r_{12}`, the distance between atoms 1 and 2,
        and :math:`a_{012}`, the angle between the three atoms.
        The rest of the array consists of all the other bond distances,
        all the other bond angles, and then all the other torsion angles.


References
----------

.. footbibliography::

"""
import logging
import warnings

import numpy as np
import copy

import MDAnalysis as mda
from .base import AnalysisBase, ResultsGroup

from MDAnalysis.lib.distances import calc_bonds, calc_angles, calc_dihedrals
from MDAnalysis.lib.mdamath import make_whole

from ..due import due, Doi

logger = logging.getLogger(__name__)


def _sort_atoms_by_mass(atoms, reverse=False):
    r"""Sorts a list of atoms by mass and then by index

    The atom index is used as a tiebreaker so that the ordering is reproducible.

    Parameters
    ----------
    ag_o : list of Atoms
        List to sort
    reverse : bool
        Atoms will be in descending order

    Returns
    -------
    ag_n : list of Atoms
        Sorted list
    """
    return sorted(atoms, key=lambda a: (a.mass, a.index), reverse=reverse)


def _find_torsions(root, atoms):
    """Constructs a list of torsion angles

    Parameters
    ----------
    root : AtomGroup
        First three atoms in the coordinate system
    atoms : AtomGroup
        Atoms that are allowed to be part of the torsion angle

    Returns
    -------
    torsions : list of AtomGroup
        list of AtomGroup objects that define torsion angles
    """
    torsions = []
    selected_atoms = list(root)
    while len(selected_atoms) < len(atoms):
        torsionAdded = False
        for a1 in selected_atoms:
            # Find a0, which is a new atom connected to the selected atom
            a0_list = _sort_atoms_by_mass(
                a
                for a in a1.bonded_atoms
                if (a in atoms) and (a not in selected_atoms)
            )
            for a0 in a0_list:
                # Find a2, which is connected to a1, is not a terminal atom,
                # and has been selected
                a2_list = _sort_atoms_by_mass(
                    a
                    for a in a1.bonded_atoms
                    if (a != a0)
                    and len(a.bonded_atoms) > 1
                    and (a in atoms)
                    and (a in selected_atoms)
                )
                for a2 in a2_list:
                    # Find a3, which is
                    # connected to a2, has been selected, and is not a1
                    a3_list = _sort_atoms_by_mass(
                        a
                        for a in a2.bonded_atoms
                        if (a != a1) and (a in atoms) and (a in selected_atoms)
                    )
                    for a3 in a3_list:
                        # Add the torsion to the list of torsions
                        torsions.append(mda.AtomGroup([a0, a1, a2, a3]))
                        # Add the new atom to selected_atoms
                        # which extends the loop
                        selected_atoms.append(a0)
                        torsionAdded = True
                        break  # out of the a3 loop
                    break  # out of the a2 loop
        if torsionAdded is False:
            print("Selected atoms:")
            print([a.index + 1 for a in selected_atoms])
            print("Torsions found:")
            print([list(t.indices + 1) for t in torsions])
            raise ValueError("Additional torsions not found.")
    return torsions


class BAT(AnalysisBase):
    """Calculate BAT coordinates for the specified AtomGroup.

    Bond-Angle-Torsions (BAT) internal coordinates will be computed for
    the group of atoms and all frame in the trajectory belonging to `ag`.

    .. versionchanged:: 2.8.0
       Enabled **parallel execution** with the ``multiprocessing`` and ``dask``
       backends; use the new method :meth:`get_supported_backends` to see all
       supported backends.

    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    @due.dcite(
        Doi("10.1002/jcc.26036"),
        description="Bond-Angle-Torsions Coordinate Transformation",
        path="MDAnalysis.analysis.bat.BAT",
    )
    def __init__(self, ag, initial_atom=None, filename=None, **kwargs):
        r"""Parameters
        ----------
        ag : AtomGroup or Universe
            Group of atoms for which the BAT coordinates are calculated.
            `ag` must have a bonds attribute.
            If unavailable, bonds may be guessed using
            :meth:`AtomGroup.guess_bonds <MDAnalysis.core.groups.AtomGroup.guess_bonds>`.
            `ag` must only include one molecule.
            If a trajectory is associated with the atoms, then the computation
            iterates over the trajectory.
        initial_atom : :class:`Atom <MDAnalysis.core.groups.Atom>`
            The atom whose Cartesian coordinates define the translation
            of the molecule. If not specified, the heaviest terminal atom
            will be selected.
        filename : str
            Name of a numpy binary file containing a saved bat array.
            If filename is not ``None``, the data will be loaded from this file
            instead of being recalculated using the run() method.

        Raises
        ------
        AttributeError
            If `ag` does not contain a bonds attribute
        ValueError
            If `ag` contains more than one molecule

        """
        super(BAT, self).__init__(ag.universe.trajectory, **kwargs)
        self._ag = ag

        # Check that the ag contains bonds
        if not hasattr(self._ag, "bonds"):
            raise AttributeError("AtomGroup has no attribute bonds")
        if len(self._ag.fragments) > 1:
            raise ValueError("AtomGroup has more than one molecule")

        # Determine the root
        # The initial atom must be a terminal atom
        terminal_atoms = _sort_atoms_by_mass(
            [a for a in self._ag.atoms if len(a.bonds) == 1], reverse=True
        )
        if initial_atom is None:
            # Select the heaviest root atoms from the heaviest terminal atom
            initial_atom = terminal_atoms[0]
        elif not initial_atom in terminal_atoms:
            raise ValueError("Initial atom is not a terminal atom")
        # The next atom in the root is bonded to the initial atom
        # Since the initial atom is a terminal atom, there is only
        # one bonded atom
        second_atom = initial_atom.bonded_atoms[0]
        # The last atom in the root is the heaviest atom
        # bonded to the second atom
        # If there are more than three atoms,
        # then the last atom cannot be a terminal atom.
        if self._ag.n_atoms != 3:
            third_atom = _sort_atoms_by_mass(
                [
                    a
                    for a in second_atom.bonded_atoms
                    if (a in self._ag)
                    and (a != initial_atom)
                    and (a not in terminal_atoms)
                ],
                reverse=True,
            )[0]
        else:
            third_atom = _sort_atoms_by_mass(
                [
                    a
                    for a in second_atom.bonded_atoms
                    if (a in self._ag) and (a != initial_atom)
                ],
                reverse=True,
            )[0]
        self._root = mda.AtomGroup([initial_atom, second_atom, third_atom])

        # Construct a list of torsion angles
        self._torsions = _find_torsions(self._root, self._ag)

        # Get indices of the root and torsion atoms
        # in a Cartesian positions array that matches the AtomGroup
        self._root_XYZ_inds = [
            (self._ag.indices == a.index).nonzero()[0][0] for a in self._root
        ]
        self._torsion_XYZ_inds = [
            [(self._ag.indices == a.index).nonzero()[0][0] for a in t]
            for t in self._torsions
        ]

        # The primary torsion is the first torsion on the list
        # with the same central atoms
        prior_atoms = [sorted([a1, a2]) for (a0, a1, a2, a3) in self._torsions]
        self._primary_torsion_indices = [
            prior_atoms.index(prior_atoms[n]) for n in range(len(prior_atoms))
        ]
        self._unique_primary_torsion_indices = list(
            set(self._primary_torsion_indices)
        )

        self._ag1 = mda.AtomGroup([ag[0] for ag in self._torsions])
        self._ag2 = mda.AtomGroup([ag[1] for ag in self._torsions])
        self._ag3 = mda.AtomGroup([ag[2] for ag in self._torsions])
        self._ag4 = mda.AtomGroup([ag[3] for ag in self._torsions])

        if filename is not None:
            self.load(filename)

    def _prepare(self):
        self.results.bat = np.zeros(
            (self.n_frames, 3 * self._ag.n_atoms), dtype=np.float64
        )

    def _single_frame(self):
        # Calculate coordinates based on the root atoms
        # The rotation axis is a normalized vector pointing from atom 0 to 1
        # It is described in two degrees of freedom
        # by the polar angle and azimuth
        if self._root.dimensions is None:
            (p0, p1, p2) = self._root.positions
        else:
            (p0, p1, p2) = make_whole(self._root, inplace=False)
        v01 = p1 - p0
        v21 = p1 - p2
        # Internal coordinates
        r01 = np.sqrt(np.einsum("i,i->", v01, v01))
        # Distance between first two root atoms
        r12 = np.sqrt(np.einsum("i,i->", v21, v21))
        # Distance between second two root atoms
        # Angle between root atoms
        a012 = np.arccos(
            max(
                -1.0,
                min(
                    1.0,
                    np.einsum("i,i->", v01, v21)
                    / np.sqrt(
                        np.einsum("i,i->", v01, v01)
                        * np.einsum("i,i->", v21, v21)
                    ),
                ),
            )
        )
        # External coordinates
        e = v01 / r01
        phi = np.arctan2(e[1], e[0])  # Polar angle
        theta = np.arccos(e[2])  # Azimuthal angle
        # Rotation to the z axis
        cp = np.cos(phi)
        sp = np.sin(phi)
        ct = np.cos(theta)
        st = np.sin(theta)
        Rz = np.array(
            [[cp * ct, ct * sp, -st], [-sp, cp, 0], [cp * st, sp * st, ct]]
        )
        pos2 = Rz.dot(p2 - p1)
        # Angle about the rotation axis
        omega = np.arctan2(pos2[1], pos2[0])
        root_based = np.concatenate((p0, [phi, theta, omega, r01, r12, a012]))

        # Calculate internal coordinates from the torsion list
        bonds = calc_bonds(
            self._ag1.positions, self._ag2.positions, box=self._ag1.dimensions
        )
        angles = calc_angles(
            self._ag1.positions,
            self._ag2.positions,
            self._ag3.positions,
            box=self._ag1.dimensions,
        )
        torsions = calc_dihedrals(
            self._ag1.positions,
            self._ag2.positions,
            self._ag3.positions,
            self._ag4.positions,
            box=self._ag1.dimensions,
        )
        # When appropriate, calculate improper torsions
        shift = torsions[self._primary_torsion_indices]
        shift[self._unique_primary_torsion_indices] = 0.0
        torsions -= shift
        # Wrap torsions to between -np.pi and np.pi
        torsions = ((torsions + np.pi) % (2 * np.pi)) - np.pi

        self.results.bat[self._frame_index, :] = np.concatenate(
            (root_based, bonds, angles, torsions)
        )

    def load(self, filename, start=None, stop=None, step=None):
        """Loads the bat trajectory from a file in numpy binary format

        Parameters
        ----------
        filename : str
            name of numpy binary file
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame

        See Also
        --------
        save: Saves the bat trajectory in a file in numpy binary format
        """
        logger.info("Choosing frames")
        self._setup_frames(self._trajectory, start, stop, step)

        logger.info("Loading file")
        self.results.bat = np.load(filename)

        # Check array dimensions
        if self.results.bat.shape != (self.n_frames, 3 * self._ag.n_atoms):
            errmsg = (
                "Dimensions of array in loaded file, "
                f"({self.results.bat.shape[0]},"
                f"{self.results.bat.shape[1]}), differ from required "
                f"dimensions of ({self.n_frames, 3*self._ag.n_atoms})"
            )
            raise ValueError(errmsg)
        # Check position of initial atom
        if (self.results.bat[0, :3] != self._root[0].position).any():
            raise ValueError(
                "Position of initial atom in file "
                + "inconsistent with current trajectory in starting frame."
            )
        return self

    def save(self, filename):
        """Saves the bat trajectory in a file in numpy binary format

        See Also
        --------
        load: Loads the bat trajectory from a file in numpy binary format
        """
        np.save(filename, self.results.bat)

    def Cartesian(self, bat_frame):
        """Conversion of a single frame from BAT to Cartesian coordinates

        One application of this function is to determine the new
        Cartesian coordinates after modifying a specific torsion angle.

        Parameters
        ----------
        bat_frame : numpy.ndarray
            an array with dimensions (3N,) with external then internal
            degrees of freedom based on the root atoms, followed by the bond,
            angle, and (proper and improper) torsion coordinates.

        Returns
        -------
        XYZ : numpy.ndarray
            an array with dimensions (N,3) with Cartesian coordinates. The first
            dimension has the same ordering as the AtomGroup used to initialize
            the class. The molecule will be whole opposed to wrapped around a
            periodic boundary.
        """
        # Split the bat vector into more convenient variables
        origin = bat_frame[:3]
        (phi, theta, omega) = bat_frame[3:6]
        (r01, r12, a012) = bat_frame[6:9]
        n_torsions = self._ag.n_atoms - 3
        bonds = bat_frame[9 : n_torsions + 9]
        angles = bat_frame[n_torsions + 9 : 2 * n_torsions + 9]
        torsions = copy.deepcopy(bat_frame[2 * n_torsions + 9 :])
        # When appropriate, convert improper to proper torsions
        shift = torsions[self._primary_torsion_indices]
        shift[self._unique_primary_torsion_indices] = 0.0
        torsions += shift
        # Wrap torsions to between -np.pi and np.pi
        torsions = ((torsions + np.pi) % (2 * np.pi)) - np.pi

        # Set initial root atom positions based on internal coordinates
        p0 = np.array([0.0, 0.0, 0.0])
        p1 = np.array([0.0, 0.0, r01])
        p2 = np.array([r12 * np.sin(a012), 0.0, r01 - r12 * np.cos(a012)])

        # Rotate the third atom by the appropriate value
        co = np.cos(omega)
        so = np.sin(omega)
        # $R_Z(\omega)$
        Romega = np.array([[co, -so, 0], [so, co, 0], [0, 0, 1]])
        p2 = Romega.dot(p2)
        # Rotate the second two atoms to point in the right direction
        cp = np.cos(phi)
        sp = np.sin(phi)
        ct = np.cos(theta)
        st = np.sin(theta)
        # $R_Z(\phi) R_Y(\theta)$
        Re = np.array(
            [[cp * ct, -sp, cp * st], [ct * sp, cp, sp * st], [-st, 0, ct]]
        )
        p1 = Re.dot(p1)
        p2 = Re.dot(p2)
        # Translate the first three atoms by the origin
        p0 += origin
        p1 += origin
        p2 += origin

        XYZ = np.zeros((self._ag.n_atoms, 3))
        XYZ[self._root_XYZ_inds[0]] = p0
        XYZ[self._root_XYZ_inds[1]] = p1
        XYZ[self._root_XYZ_inds[2]] = p2

        # Place the remaining atoms
        for (a0, a1, a2, a3), r01, angle, torsion in zip(
            self._torsion_XYZ_inds, bonds, angles, torsions
        ):

            p1 = XYZ[a1]
            p3 = XYZ[a3]
            p2 = XYZ[a2]

            sn_ang = np.sin(angle)
            cs_ang = np.cos(angle)
            sn_tor = np.sin(torsion)
            cs_tor = np.cos(torsion)

            v21 = p1 - p2
            v21 /= np.sqrt(np.einsum("i,i->", v21, v21))
            v32 = p2 - p3
            v32 /= np.sqrt(np.einsum("i,i->", v32, v32))

            vp = np.cross(v32, v21)
            cs = np.einsum("i,i->", v21, v32)

            sn = max(np.sqrt(1.0 - cs * cs), 0.0000000001)
            vp = vp / sn
            vu = np.cross(vp, v21)

            XYZ[a0] = p1 + r01 * (
                vu * sn_ang * cs_tor + vp * sn_ang * sn_tor - v21 * cs_ang
            )
        return XYZ

    @property
    def atoms(self):
        """The atomgroup for which BAT are computed (read-only property)"""
        return self._ag

    def _get_aggregator(self):
        return ResultsGroup(lookup={"bat": ResultsGroup.ndarray_vstack})



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_allclose
import pytest

import MDAnalysis as mda
from MDAnalysis.lib import qcprot

from MDAnalysisTests.datafiles import PSF, DCD


@pytest.mark.parametrize("dtype", [np.float64, np.float32])
class TestQCProt:
    def test_dummy(self, dtype):
        a = np.array([[1.0, 1.0, 2.0]], dtype=dtype)
        b = np.array([[1.0, 1.0, 1.0]], dtype=dtype)
        rot = np.zeros(9, dtype=dtype)

        r = qcprot.CalcRMSDRotationalMatrix(a, b, 1, rot, None)

        err = 0.0001 if dtype is np.float32 else 0.000001

        assert r == pytest.approx(0.7174389197644676, abs=err)
        assert rot.dtype == dtype
        assert_allclose(rot.reshape(3, 3), np.eye(3))

    def test_regression(self, dtype):
        u = mda.Universe(PSF, DCD)
        prot = u.select_atoms("protein")
        weights = prot.masses.astype(dtype)
        weights /= np.mean(weights)
        p1 = prot.positions.astype(dtype)
        u.trajectory[1]
        p2 = prot.positions.astype(dtype)
        rot = np.zeros(9, dtype=dtype)

        r = qcprot.CalcRMSDRotationalMatrix(p1, p2, len(prot), rot, weights)

        rot_ref = np.array(
            [
                0.999998,
                0.001696,
                0.001004,
                -0.001698,
                0.999998,
                0.001373,
                -0.001002,
                -0.001375,
                0.999999,
            ],
            dtype=dtype,
        )

        err = 0.001 if dtype is np.float32 else 0.000001
        assert r == pytest.approx(0.6057544485785074, abs=err)
        assert_allclose(rot, rot_ref, atol=err)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""Core functions of MDAnalysis
============================

The basic class is an :class:`~MDAnalysis.core.groups.AtomGroup`; the whole
simulation is called the
:class:`~MDAnalysis.core.universe.Universe`. Selections are computed on an
:class:`~MDAnalysis.core.groups.AtomGroup` and return another
:class:`~MDAnalysis.core.groups.AtomGroup`.

To get started, load the Universe::

  u = Universe(topology_file, trajectory_file)

A simple selection of all water oxygens within 4 A of the protein::

  water_shell = u.select_atoms('name OH2 and around 4.0 protein')
  water_shell.n_atoms           # how many waters were selected
  water_shell.total_mass()       # their total mass

:class:`~MDAnalysis.core.groups.AtomGroup` instances have various methods that
allow calculation of simple properties. For more complicated analysis, obtain
the coordinates as a numpy array ::

  coords = water_shell.positions

and write your own Python code.
"""

__all__ = ["AtomGroup", "Selection"]

from .groups import AtomGroup
from .selection import Selection



import MDAnalysis as mda
import numpy as np
from MDAnalysis.coordinates.GRO import GROReader
from MDAnalysis.topology.GROParser import GROParser
from MDAnalysisTests.datafiles import GRO


class GROReadBench(object):
    def time_read_GRO_coordinates(self):
        """Benchmark reading of standard testsuite GRO file."""
        GROReader(GRO)

    def time_parse_GRO_file(self):
        """Time to create topology from GRO file"""
        p = GROParser(GRO)
        top = p.parse()

    def time_create_GRO_universe(self):
        """Time to create MDA Universe of GRO"""
        u = mda.Universe(GRO)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Cluster representation --- :mod:`MDAnalysis.analysis.encore.clustering.ClusterCollection`
=========================================================================================

The module contains the Cluster and ClusterCollection classes which are
designed to store results from clustering algorithms.

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

"""
import numpy as np


class Cluster(object):
    """
    Generic Cluster class for clusters with centroids.

    Attributes
    ----------

    id : int
        Cluster ID number. Useful for the ClustersCollection class

    metadata : iterable
        dict of lists or numpy.array, containing metadata for the cluster
        elements. The iterable must return the same number of elements as
        those that belong to the cluster.

    size : int
        number of elements.

    centroid : element object
        cluster centroid.

    elements : numpy.array
        array containing the cluster elements.
    """

    def __init__(self, elem_list=None, centroid=None, idn=None, metadata=None):
        """Class constructor. If elem_list is None, an empty cluster is created
            and the remaining arguments ignored.

        Parameters
        ----------

        elem_list : numpy.array or None
            numpy array of cluster elements

        centroid : None or element object
            centroid

        idn : int
            cluster ID

        metadata : iterable
            metadata, one value for each cluster element. The iterable
            must have the same length as the elements array.

        """

        self.id = idn

        if elem_list is None:
            self.size = 0
            self.elements = np.array([])
            self.centroid = None
            self.metadata = {}
            return

        self.metadata = {}
        self.elements = elem_list
        if centroid not in self.elements:
            raise LookupError(
                "Centroid of cluster not found in the element list"
            )

        self.centroid = centroid
        self.size = self.elements.shape[0]
        if metadata:
            for name, data in metadata.items():
                if len(data) != self.size:
                    raise TypeError(
                        'Size of metadata having label "{0}" '
                        "is not equal to the number of cluster "
                        "elements".format(name)
                    )
            self.add_metadata(name, data)

    def __iter__(self):
        """
        Iterate over elements in cluster
        """
        return iter(self.elements)

    def __len__(self):
        """
        Size of cluster
        """
        return len(self.elements)

    def add_metadata(self, name, data):
        if len(data) != self.size:
            raise TypeError(
                "Size of metadata is not equal to the number of "
                "cluster elements"
            )
        self.metadata[name] = np.array(data)

    def __repr__(self):
        """
        Textual representation
        """
        if self.size == 0:
            return "<Cluster with no elements>"
        else:
            return "<Cluster with {0} elements, centroid={1}, id={2}>".format(
                self.size, self.centroid, self.id
            )


class ClusterCollection(object):
    """Clusters collection class; this class represents the results of a full
    clustering run. It stores a group of clusters defined as
    encore.clustering.Cluster objects.

    Attributes
    ----------

    clusters : list
        list of of Cluster objects which are part of the Cluster collection

    """

    def __init__(self, elements=None, metadata=None):
        """Class constructor. If elements is None, an empty cluster collection
        will be created. Otherwise, the constructor takes as input an
        iterable of ints, for instance:

        [ a, a, a, a, b, b, b, c, c, ... , z, z ]

        the variables a,b,c,...,z are cluster centroids, here as cluster
        element numbers (i.e. 3 means the 4th element of the ordered input
        for clustering). The array maps a correspondence between
        cluster elements (which are implicitly associated with the
        position in the array) with centroids, i. e. defines clusters.
        For instance:

        [ 1, 1, 1, 4, 4, 5 ]

        means that elements 0, 1, 2 form a cluster which has 1 as centroid,
        elements 3 and 4 form a cluster which has 4 as centroid, and
        element 5 has its own cluster.


        Parameters
        ----------

        elements : iterable of ints or None
            clustering results. See the previous description for details

        metadata : {str:list, str:list,...} or None
            metadata for the data elements. The list must be of the same
            size as the elements array, with one value per element.

        """
        idn = 0
        if elements is None:
            self.clusters = None
            return

        if not len(set((type(el) for el in elements))) == 1:
            raise TypeError("all the elements must have the same type")
        self.clusters = []
        elements_array = np.array(elements)
        centroids = np.unique(elements_array)
        for i in centroids:
            if elements[i] != i:
                raise ValueError(
                    "element {0}, which is a centroid, doesn't "
                    "belong to its own cluster".format(elements[i])
                )
        for c in centroids:
            this_metadata = {}
            this_array = np.where(elements_array == c)
            if metadata:
                for k, v in metadata.items():
                    this_metadata[k] = np.asarray(v)[this_array]
            self.clusters.append(
                Cluster(
                    elem_list=this_array[0],
                    idn=idn,
                    centroid=c,
                    metadata=this_metadata,
                )
            )

            idn += 1

    def get_ids(self):
        """
        Get the ID numbers of the clusters

        Returns
        -------

        ids : list of int
        list of cluster ids
        """
        return [v.id for v in self.clusters]

    def get_centroids(self):
        """
        Get the centroids of the clusters

        Returns
        -------

        centroids : list of cluster element objects
        list of cluster centroids
        """

        return [v.centroid for v in self.clusters]

    def __iter__(self):
        """
        Iterate over clusters

        """
        return iter(self.clusters)

    def __len__(self):
        """
        Length of clustering collection
        """
        return len(self.clusters)

    def __repr__(self):
        """
        Textual representation
        """
        if self.clusters is None:
            return "<ClusterCollection with no clusters>"
        else:
            return "<ClusterCollection with {0} clusters>".format(
                len(self.clusters)
            )
